/*
 * Created on November 9, 2005, 3:37 PM
 *
 */

package com.sunlabs.archsim.fleet.ships.adder_oob;

import com.sunlabs.archsim.base.hierarchy.CompositeEntity;
import com.sunlabs.archsim.base.simulation.Command;
import com.sunlabs.archsim.base.simulation.Delay;
import com.sunlabs.archsim.fleet.common.Data;
import com.sunlabs.archsim.fleet.common.Payload;
import com.sunlabs.archsim.gasp.Action;
import com.sunlabs.archsim.gasp.terminals.InputTerminal;
import com.sunlabs.archsim.gasp.Function;
import com.sunlabs.archsim.gasp.State;
import com.sunlabs.archsim.gasp.Terminal;
import com.sunlabs.archsim.gasp.terminals.OutputTerminal;

/*
 * Copyright � 2005 Sun Microsystems Inc. All rights reserved. Use is subject to license terms.
 *
 *
 * @author mh193489
 */
public class AdderAction extends Action {
    private Delay adderDelay= AdderDefaults.ADDER_DELAY_DEF;
    
    Function outFunction= new AddData();
    Function ackInFunction= new AckFunction();
    
    InputTerminal in1;
    InputTerminal in2;
    InputTerminal ackOut;
    
    OutputTerminal out, ackIn1, ackIn2;
    
    Data data1;
    Data data2;
    
    private boolean activeAckOut, activeIn1, activeIn2;
    
    public AdderAction(String name, CompositeEntity parent) {
        super(name, parent);
        outputFunctions.put(AdderDefaults.OUT, outFunction);
        outputFunctions.put(AdderDefaults.ACK_IN1, ackInFunction);
        outputFunctions.put(AdderDefaults.ACK_IN2, ackInFunction);
        
        notifyStateIsActiveCommand= new StateActiveCommand();
        notifyStateIsInactiveCommand= new StateInactiveCommand();    
    } // AdderAction
    
    public void setDelay(Delay d) {
        adderDelay= (Delay)d.clone();
        outFunction.setDelay(adderDelay);
        ackInFunction.setDelay(adderDelay);
    }
    
    public boolean selfCheck() {
        boolean OK= true;
        boolean check;
        
        check= (this.getTerminal(AdderDefaults.IN1) != null);
        OK= OK && check;
        if (!check) {
            director.error("Action " + getAlias() + ", id= " + getID()
            + ", does not have attached terminal in1");
        } else {
            in1= (InputTerminal)this.getTerminal(AdderDefaults.IN1);
        }
        
        check= (this.getTerminal(AdderDefaults.ACK_IN1) != null);
        OK= OK && check;
        if (!check) {
            director.error("Action " + getAlias() + ", id= " + getID()
            + ", does not have attached terminal ackIn1");
        } else {
            ackIn1= (OutputTerminal)this.getTerminal(AdderDefaults.ACK_IN1);
        }
        
        check= (this.getTerminal(AdderDefaults.IN2) != null);
        OK= OK && check;
        if (!check) {
            director.error("Action " + getAlias() + ", id= " + getID()
            + ", does not have attached terminal in2");
        } else {
            in2= (InputTerminal)this.getTerminal(AdderDefaults.IN2);
        }
        
        check= (this.getTerminal(AdderDefaults.ACK_IN2) != null);
        OK= OK && check;
        if (!check) {
            director.error("Action " + getAlias() + ", id= " + getID()
            + ", does not have attached terminal ackIn2");
        } else {
            ackIn2= (OutputTerminal)this.getTerminal(AdderDefaults.ACK_IN2);
        }

        
        check= (this.getTerminal(AdderDefaults.OUT) != null);
        OK= OK && check;
        if (!check) {
            director.error("Action " + getAlias() + ", id= " + getID()
            + ", does not have attached terminal out");
        } else {
            out= (OutputTerminal)this.getTerminal(AdderDefaults.OUT);
        }
        
        check= (this.getTerminal(AdderDefaults.ACK_OUT) != null);
        OK= OK && check;
        if (!check) {
            director.error("Action " + getAlias() + ", id= " + getID()
            + ", does not have attached terminal ackOut");
        } else {
            ackOut= (InputTerminal)this.getTerminal(AdderDefaults.ACK_OUT);
        }
        
        // also run the self check on the superclass
        check= super.selfCheck();
        OK= OK && check;
        
        return OK;
    }
    
    private void tryToFire() {
        if (activeAckOut && activeIn1 && activeIn2) {
            data1= ((Payload)in1.getData()).data;
            data2= ((Payload)in2.getData()).data;
            
            if ( ((data1.value != null) && (data2.value != null)) || (!data1.valid && (data1.value == null) && !data2.valid && (data2.value == null)) ) {
                in1.setStateInactive();
                in2.setStateInactive();
                ackOut.setStateInactive();
                computeAndFire(out);
                computeAndFire(ackIn1);
                computeAndFire(ackIn2);                
    
                
            } else if (data2.valid && (data2.value != null)) {
//                director.error("ADDER: consuming and ignoring in2");
                // ignore in2
                in2.setStateInactive();
                computeAndFire(ackIn2);
                
            } else if (data1.valid && (data1.value != null)) {
//                director.error("ADDER: consuming and ignoring in1");
                // ignore in1
                in1.setStateInactive();
                computeAndFire(ackIn1);
            } else {
//                director.error("ADDER: UNNOWN combination of inputs");
            }
        }
    }
    
    protected class StateActiveCommand extends Command {
        public void execute(Object param) {
            // update the count of active inputs.
            // note that the state itself contains a check
            // whether it became active twice without being
            // inactive in between
            Terminal src= (Terminal)param;
            if (journalAction) {
                State state= src.getState();
                director.journal("Action " + getAlias()
                + ", id= " + getID() + " notifed that state "
                        + state.getAlias() + ", id= " + state.getID()
                        + " is active");
            }
            
            
            if (src == in1) {
//                director.error("Activated: in (" + src.getState().getName() + ")");
                
                activeIn1= true;
                
                tryToFire();
                
            } else if (src == ackIn1) {
//                director.error("Activated: ackIn1 (" + src.getState().getName() + ")");
                
            } else if (src == in2) {
//                director.error("Activated: in2 (" + src.getState().getName() + ")");
                
                activeIn2= true;
                
                tryToFire();
                
                
            } else if (src == ackIn2) {
//                director.error("Activated: ackIn2 (" + src.getState().getName() + ")");
                
            } else if (src == out) {
//                director.error("Activated: out (" + src.getState().getName() + ")");
                
            } else if (src == ackOut) {
//                director.error("Activated: ackOut (" + src.getState().getName() + ")");
                
                activeAckOut= true;
                tryToFire();
                
            } else{
//                director.error("Activated: UNKNOWN");
            }
            
        } // execute
    } // StateActiveCommand
    
    
    protected class StateInactiveCommand extends Command {
        public void execute(Object param) {
            Terminal src= (Terminal)param;
            
            if (src == in1) {
//                director.error("Deactivated: in1 (" + src.getState().getName() + ")");
                
                activeIn1= false;
                
            } else if (src == ackIn1) {
//                director.error("Deactivated: ackIn1 (" + src.getState().getName() + ")");
                
            } else if (src == in2) {
//                director.error("Deactivated: in2 (" + src.getState().getName() + ")");
                
                activeIn2= false;
                
            } else if (src == ackIn2) {
//                director.error("Deactivated: ackIn2 (" + src.getState().getName() + ")");
                
            } else if (src == out) {
//                director.error("Deactivated: out (" + src.getState().getName() + ")");
                
            } else if (src == ackOut) {
//                director.error("Deactivated: ackOut (" + src.getState().getName() + ")");
                
                activeAckOut= false;
                
                
            } else {
//                director.error("Deactivated: UNKNOWN (" + src.getState().getName() + ")");
            }
            
            
            if (journalAction) {
                State state= src.getState();
                director.journal("Action " + getAlias()
                + ", id= " + getID() + " notifed that state "
                        + state.getAlias() + ", id= " + state.getID()
                        + " is INactive");
            }
        } // execute
    } // StateInactiveCommand
    
    
    private class AddData extends Function {
        
        public AddData() {
            super(AdderDefaults.ADD_DATA);
            functionDelay= adderDelay;
        }
        
        public Object value() {
            // get data from input terminals
            
            // check if input data is valid
            if(data1.valid && data2.valid) {
                // both inputs valid, hence compute the sum
                Integer sum= (Integer)data1.value + (Integer)data2.value;
                //check for over or underflow
                if (sum <= AdderDefaults.UNDERFLOW) {
                    return new Data(AdderDefaults.MINUS_INFINITY,false);
                    
                } else if (AdderDefaults.OVERFLOW <= sum) {
//                    director.error("ADDER OVERFLOW: " + data1.value + " + " + data2.value + " = " + sum + " >= " + AdderDefaults.OVERFLOW );
                    return new Data(AdderDefaults.PLUS_INFINITY, false);
                    
                } else {
                    return new Data(sum, true);
                }
            } else if (!data1.valid && !data2.valid) {
                return new Data(null, false);
                
            } else {
                // at least one input invalid, hence return invalid sum
                return new Data(AdderDefaults.OOB, false);
            }
        } // value
    } // AddData
    
}


